#include "buffer.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

void obuf_init(obuf *o, unsigned long memsize, unsigned long frames_per_buf)
{
  unsigned long nb_buffers = memsize / (frames_per_buf * 4);
  if (nb_buffers < 2) { printf("not enough buffer in obuf\n"); exit(1); }
  o->data = malloc(frames_per_buf * 4 * nb_buffers);
  if(!o->data){printf("oom\n");exit(1);}
  o->play_next = 0;
  o->putpos = 0;
  o->len = nb_buffers;
  o->size = frames_per_buf * 4 * nb_buffers;
  o->count = 0;
  o->partial = 0;
  o->store_count = 0;
  o->buffer_size = frames_per_buf * 4;
}

int obuf_hasroom(obuf *o)
{
  /* at least 2 buffers must be free, not sure about this partial stuff... */
  return o->count < o->len - 1;
}

int obuf_isempty(obuf *o)
{
  return o->count == 0;
}

void obuf_put(obuf *o, unsigned char *f, int len)
{
  int slen = len;
  int l;
  o->store_count += len / 4;
  while (len) {
    l = len;
    if (o->putpos + l > o->size) l = o->size - o->putpos;
    memcpy(o->data + o->putpos, f, l);
    o->putpos += l; if (o->putpos == o->size) o->putpos = 0;
    len -= l;
    f += l;
  }
  o->partial += slen;
  if (o->partial >= o->buffer_size) {
    o->count++;
    o->partial -= o->buffer_size;
  }
}

unsigned char *obuf_play(obuf *o)
{
  unsigned char *ret = o->data + o->play_next * o->buffer_size;
  o->play_next++; if (o->play_next == o->len) o->play_next = 0;
  o->count--;
  return ret;
}

void obuf_fill(obuf *o)
{
  if (!o->partial) return;
  o->store_count += (o->buffer_size - o->partial) / 4;
  while (o->partial != o->buffer_size) {
    o->data[o->putpos] = 0;
    o->partial++;
    o->putpos++;
  }
  o->count++;
}
