#include "data.h"
#include "oss.h"
#include "alsa.h"
#include "buffer.h"
#include "position.h"
#include "skipper.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/select.h>
#include <errno.h>
#include <string.h>

static unsigned long pos;

void clear_song_pos(void)
{
  pos = 0;
}

static unsigned long select_count;
static unsigned long silence_count;

#ifdef ALSA
/* return:
 * 1 - alsa ready
 * 2 - pipefd ready
 * 3 - both
 * pass pipefd=-1 to not test it in select
 */
static int do_wait(int pipefd)
{
  fd_set rd;
  fd_set wr;
  fd_set er;
  int ret = 0;
  int n;
  int i;
  int max = pipefd;
  extern int alsa_fd_rd_nb;
  extern int alsa_fd_wr_nb;
  extern int alsa_fd_rd[10];
  extern int alsa_fd_wr[10];

  for (i = 0; i < alsa_fd_rd_nb; i++)
    if (max < alsa_fd_rd[i]) max = alsa_fd_rd[i];
  for (i = 0; i < alsa_fd_wr_nb; i++)
    if (max < alsa_fd_wr[i]) max = alsa_fd_wr[i];

  do {
    FD_ZERO(&rd);
    FD_ZERO(&wr);
    FD_ZERO(&er);
    if (pipefd != -1) {
      FD_SET(pipefd, &rd);
      FD_SET(pipefd, &er);
    }
    for (i = 0; i < alsa_fd_rd_nb; i++) {
      FD_SET(alsa_fd_rd[i], &rd);
      FD_SET(alsa_fd_rd[i], &er);
    }
    for (i = 0; i < alsa_fd_wr_nb; i++) {
      FD_SET(alsa_fd_wr[i], &wr);
      FD_SET(alsa_fd_wr[i], &er);
    }
    n = select(max+1, &rd, &wr, &er, NULL); select_count++;
  } while (n == -1 && errno == EINTR);
  if (n == -1) {printf("select error\n"); exit(1); }
  for (i = 0; i < alsa_fd_rd_nb; i++) if (FD_ISSET(alsa_fd_rd[i], &er))
    {printf("%d:error\n",__LINE__); exit(1); }
  for (i = 0; i < alsa_fd_wr_nb; i++) if (FD_ISSET(alsa_fd_wr[i], &er))
    {printf("%d:error\n",__LINE__); exit(1); }
  if (FD_ISSET(pipefd, &er)) {printf("%d:error\n",__LINE__); exit(1); }
  if (alsa_can_play(&rd, &wr)) ret |= 1;
  if (pipefd != -1 && FD_ISSET(pipefd, &rd)) ret |= 2;
  return ret;
}
#endif /* ALSA */

volatile int vol;
volatile float volr;

static char *silence;
static unsigned char *buf;
static void output(char *buf, int len, obuf *b)
{
  int i;
  unsigned long h, m, s;
  static unsigned long oh = 0, om = 0, os = 61;

  if (len%4) { fprintf(stderr, "fatal error buddy\n"); exit(1); }

  s = pos / 44100;
  h = s / (60 * 60);
  s -= h * 60 * 60;
  m = s / 60;
  s -= m * 60;

  if (h != oh || m != om || s != os) {
    extern int pid;
    oh = h;
    om = m;
    os = s;
#ifdef ALSA
#if 0
    printf("T %2.2ld:%2.2ld:%2.2ld (%d %d %d %d %d) sc %lu  skip silence: kill -usr2 %d\r",
           h, m, s, b->putpos, b->play_next, b->size, b->count, b->partial, select_count, pid);
    select_count = 0;
#else
    printf(
    "T %2.2ld:%2.2ld:%2.2ld (sel %.4lu sil %lu) skip silence: kill -usr2 %d \r",
           h, m, s, select_count, silence_count, pid);
    select_count = 0;
#endif
#else
    printf("T %2.2ld:%2.2ld:%2.2ld  skip silence: kill -usr2 %d\r",
           h, m, s, pid);
#endif
    fflush(stdout);
  }

  if (buf != silence) {
    pos += len / 4;
    advance_position(len / 4);
  }

  if (skip_silence) {
    short *sg = (short *)buf;
    double energy = 0;
    for (i = 0; i < len / 2; i++)
      energy += sg[i] * sg[i];
    energy /= len/2;
    if (energy < 100) return;
    skip_silence = 0;
  }

  if (vol) {
    short *sg = (short *)buf;
    for (i = 0; i < len / 2; i++) {
      float v = sg[i] * volr;
      if (v < -32768) v = -32768;
      if (v > 32767) v = 32767;
      sg[i] = v;
    }
  }

#ifndef ALSA
  if (!play_oss(buf, len)) exit(1);
#else
  play_alsa(buf, len);
#endif
}

static void ffmpeg(const char *file, obuf *b, int file_number)
{
  int pipefd[2];
  int son;
  int i;
  int n;
  int w;
  skipper s;

  if (pipe(pipefd) == -1) { perror("pipe"); exit(1); }

  son = fork(); if (son == -1) { perror("fork"); exit(1); }

  if (!son) {
    close(pipefd[0]);
    if (dup2(pipefd[1], 1) == -1) { perror("dup2"); exit(1); }
    close(2);
    execlp("ffmpeg", "ffmpeg", "-i", file, "-f", "rawvideo", "-vn",
           "-ac", "2", "-ar", "44100", "-acodec", "pcm_s16le", "-",
           (char*)NULL);
    perror("execl");
    exit(0);
  }

  init_skipper(&s, file, b->buffer_size);

  close(pipefd[1]);
  while (1) {
#ifdef ALSA
    w = do_wait(obuf_hasroom(b) ? pipefd[0] : -1);
    if (w & 1) {
      if (!obuf_isempty(b))
        output((char *)obuf_play(b), b->buffer_size, b);
      else {
        if (skip_silence || (file_number == 0 && pos == 0)) {
          output(silence, b->buffer_size, b);
          silence_count++;
        }
      }
    }
    if (w & 2) {
      n = read(pipefd[0], buf, b->buffer_size); if (n <= 0) break;
      if (n&3) {printf("read from pipe error\n"); exit(1); }
      for (i = 0; i < n / 4; i++) {
        short dl, dr;
        short ol, or;
        dl = buf[i*4] + ((int)(buf[i*4+1]) << 8);
        dr = buf[i*4+2] + ((int)(buf[i*4+3]) << 8);
        filter(dl, dr, &ol, &or);
        buf[i*4] = ol & 255;
        buf[i*4+1] = (ol>>8) & 255;
        buf[i*4+2] = or & 255;
        buf[i*4+3] = (or>>8) & 255;
      }
      //obuf_put(b, buf, n);
      feed_skipper(&s, buf, n, b);
    }
#else
    n = read(pipefd[0], buf, b->buffer_size);
    if (!n) break;
    output(buf, n, b);
#endif
  }
  flush_skipper(&s, b);
  free_skipper(&s);
  close(pipefd[0]);
  wait(NULL);
}

void play(playlist *l)
{
  obuf b;
  int i;
  int w;

#ifdef ALSA
  {
    extern unsigned long frames_per_period;
    obuf_init(&b, 1024 * 1024, frames_per_period);
  }
#else
  obuf_init(&b, 1024 * 1024, 512);
#endif

  buf = malloc(b.buffer_size);
  silence = malloc(b.buffer_size);
  if (!buf || !silence) { printf("oom\n"); exit(1); }
  memset(silence, 0, b.buffer_size);

  for (i = 0; i < l->size; i++) {
    position_next_file(l->list[i].name, b.store_count);
    ffmpeg(l->list[i].name, &b, i);
  }

  obuf_fill(&b);

  while (!obuf_isempty(&b)) {
#ifdef ALSA
    w = do_wait(-1);
    if (w & 1) output((char *)obuf_play(&b), b.buffer_size, &b);
#else
    output((char *)obuf_play(&b), b->buffer_size, &b);
#endif
  }

  free(buf);
  free(silence);
}
